@@ -3,53 +3,106 @@
 """Module implementation."""
 
 import gzip
+import json
 import logging
+import lzma
 import os
-import pickle
-import warnings
+import tarfile
+import zipfile
 from contextlib import contextmanager
 from pathlib import Path
-from typing import Any, Mapping, Optional, Sequence, TYPE_CHECKING, Union
+from typing import TYPE_CHECKING, Any, Mapping, Optional, Sequence, Union
 
+from . import utils
+from .constants import (
+    PYSTOW_HOME_ENVVAR,
+    PYSTOW_NAME_DEFAULT,
+    PYSTOW_NAME_ENVVAR,
+    PYSTOW_USE_APPDIRS,
+    README_TEXT,
+)
 from .utils import (
-    download, download_from_google, download_from_s3, getenv_path, mkdir, name_from_s3_key, name_from_url, read_rdf,
-    read_tarfile_csv, read_zipfile_csv,
+    download_from_google,
+    download_from_s3,
+    getenv_path,
+    mkdir,
+    name_from_s3_key,
+    name_from_url,
+    read_rdf,
+    read_tarfile_csv,
+    read_tarfile_xml,
+    read_zipfile_csv,
 )
 
+try:
+    import pickle5 as pickle
+except ImportError:
+    import pickle  # type:ignore
+
 if TYPE_CHECKING:
-    import rdflib
-    import pandas as pd
     import botocore.client
+    import pandas as pd
+    import rdflib
 
 logger = logging.getLogger(__name__)
 
-PYSTOW_NAME_ENVVAR = 'PYSTOW_NAME'
-PYSTOW_HOME_ENVVAR = 'PYSTOW_HOME'
-PYSTOW_NAME_DEFAULT = '.data'
-
 
 def get_name() -> str:
     """Get the PyStow home directory name."""
     return os.getenv(PYSTOW_NAME_ENVVAR, default=PYSTOW_NAME_DEFAULT)
 
 
+def _use_appdirs() -> bool:
+    return os.getenv(PYSTOW_USE_APPDIRS) in {"true", "True"}
+
+
 def get_home(ensure_exists: bool = True) -> Path:
     """Get the PyStow home directory."""
-    default = Path.home() / get_name()
+    if _use_appdirs():
+        from appdirs import user_data_dir
+
+        default = Path(user_data_dir())
+    else:
+        default = Path.home() / get_name()
     return getenv_path(PYSTOW_HOME_ENVVAR, default, ensure_exists=ensure_exists)
 
 
 def get_base(key: str, ensure_exists: bool = True) -> Path:
-    """Get the base directory for a module."""
+    """Get the base directory for a module.
+
+    :param key:
+        The name of the module. No funny characters. The envvar
+        <key>_HOME where key is uppercased is checked first before using
+        the default home directory.
+    :param ensure_exists:
+        Should all directories be created automatically?
+        Defaults to true.
+    :returns:
+        The path to the given
+    """
     _assert_valid(key)
-    envvar = f'{key.upper()}_HOME'
-    default = get_home(ensure_exists=False) / key
+    envvar = f"{key.upper()}_HOME"
+    if _use_appdirs():
+        from appdirs import user_data_dir
+
+        default = Path(user_data_dir(appname=key))
+    else:
+        default = get_home(ensure_exists=False) / key
     return getenv_path(envvar, default, ensure_exists=ensure_exists)
 
 
 def _assert_valid(key: str) -> None:
-    if '.' in key:
-        raise ValueError(f'The module should not have a dot in it: {key}')
+    if "." in key:
+        raise ValueError(f"The module should not have a dot in it: {key}")
+
+
+def ensure_readme():
+    """Ensure there's a README in the PyStow data directory."""
+    readme_path = get_home(ensure_exists=True).joinpath("README.md")
+    if readme_path.is_file():
+        return
+    with readme_path.open("w", encoding="utf8") as file:
+        print(README_TEXT, file=file)  # noqa:T001
 
 
 class Module:
@@ -68,15 +121,29 @@ class Module:
         mkdir(self.base, ensure_exists=ensure_exists)
 
     @classmethod
-    def from_key(cls, key: str, *subkeys: str, ensure_exists: bool = True) -> 'Module':
-        """Get a module for the given directory or one of its subdirectories."""
+    def from_key(cls, key: str, *subkeys: str, ensure_exists: bool = True) -> "Module":
+        """Get a module for the given directory or one of its subdirectories.
+
+        :param key:
+            The name of the module. No funny characters. The envvar
+            <key>_HOME where key is uppercased is checked first before using
+            the default home directory.
+        :param subkeys:
+            A sequence of additional strings to join. If none are given,
+            returns the directory for this module.
+        :param ensure_exists:
+            Should all directories be created automatically?
+            Defaults to true.
+        :return:
+            A module
+        """
         base = get_base(key, ensure_exists=False)
         rv = cls(base=base, ensure_exists=ensure_exists)
         if subkeys:
             rv = rv.submodule(*subkeys, ensure_exists=ensure_exists)
         return rv
 
-    def submodule(self, *subkeys: str, ensure_exists: bool = True) -> 'Module':
+    def submodule(self, *subkeys: str, ensure_exists: bool = True) -> "Module":
         """Get a module for a subdirectory of the current module.
 
         :param subkeys:
@@ -118,10 +185,17 @@ class Module:
             rv = rv.joinpath(name)
         return rv
 
-    def get(self, *args, **kwargs):
-        """Get a subdirectory of the current module, deprecated in favor of :meth:`join`."""
-        warnings.warn('Use Module.join instead of Module.get', DeprecationWarning)
-        return self.join(*args, **kwargs)
+    def joinpath_sqlite(self, *subkeys: str, name: str) -> str:
+        """Get an SQLite database connection string.
+
+        :param subkeys:
+            A sequence of additional strings to join. If none are given,
+            returns the directory for this module.
+        :param name: The name of the database file.
+        :return: A SQLite path string.
+        """
+        path = self.join(*subkeys, name=name, ensure_exists=True)
+        return f"sqlite:///{path.as_posix()}"
 
     def ensure(
         self,
@@ -151,7 +225,7 @@ class Module:
         if name is None:
             name = name_from_url(url)
         path = self.join(*subkeys, name=name, ensure_exists=True)
-        download(
+        utils.download(
             url=url,
             path=path,
             force=force,
@@ -159,6 +233,32 @@ class Module:
         )
         return path
 
+    def ensure_untar(
+        self,
+        *subkeys: str,
+        url: str,
+        name: Optional[str] = None,
+        directory: Optional[str] = None,
+        force: bool = False,
+        download_kwargs: Optional[Mapping[str, Any]] = None,
+        extract_kwargs: Optional[Mapping[str, Any]] = None,
+    ) -> Path:
+        """Ensure a tar file is downloaded and unarchived."""
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
+        if directory is None:
+            # rhea-rxn.tar.gz -> rhea-rxn
+            suffixes_len = sum(len(suffix) for suffix in path.suffixes)
+            directory = path.name[:-suffixes_len]
+        unzipped_path = path.parent.joinpath(directory)
+        if unzipped_path.is_dir() and not force:
+            return unzipped_path
+        unzipped_path.mkdir(exist_ok=True, parents=True)
+        with tarfile.open(path) as tar_file:
+            tar_file.extractall(unzipped_path, **(extract_kwargs or {}))
+        return unzipped_path
+
     @contextmanager
     def ensure_open(
         self,
@@ -167,16 +267,78 @@ class Module:
         name: Optional[str] = None,
         force: bool = False,
         download_kwargs: Optional[Mapping[str, Any]] = None,
-        mode: str = 'r',
+        mode: str = "r",
         open_kwargs: Optional[Mapping[str, Any]] = None,
     ):
         """Ensure a file is downloaded then open it."""
-        path = self.ensure(*subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs)
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
         open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
-        open_kwargs.setdefault('mode', mode)
+        open_kwargs.setdefault("mode", mode)
         with path.open(**open_kwargs) as file:
             yield file
 
+    @contextmanager
+    def ensure_open_lzma(
+        self,
+        *subkeys: str,
+        url: str,
+        name: Optional[str] = None,
+        force: bool = False,
+        download_kwargs: Optional[Mapping[str, Any]] = None,
+        mode: str = "rt",
+        open_kwargs: Optional[Mapping[str, Any]] = None,
+    ):
+        """Ensure a file is downloaded then open it with :mod:`lzma`."""
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
+        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
+        open_kwargs.setdefault("mode", mode)
+        with lzma.open(path, **open_kwargs) as file:
+            yield file
+
+    @contextmanager
+    def ensure_open_tarfile(
+        self,
+        *subkeys: str,
+        url: str,
+        inner_path: str,
+        name: Optional[str] = None,
+        force: bool = False,
+        download_kwargs: Optional[Mapping[str, Any]] = None,
+    ):
+        """Ensure a file is downloaded then open it with :mod:`tarfile`."""
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
+        with tarfile.open(path) as tar_file:
+            with tar_file.extractfile(inner_path) as file:  # type:ignore
+                yield file
+
+    @contextmanager
+    def ensure_open_zip(
+        self,
+        *subkeys: str,
+        url: str,
+        inner_path: str,
+        name: Optional[str] = None,
+        force: bool = False,
+        download_kwargs: Optional[Mapping[str, Any]] = None,
+        mode: str = "r",
+        open_kwargs: Optional[Mapping[str, Any]] = None,
+    ):
+        """Ensure a file is downloaded then open it with :mod:`zipfile`."""
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
+        open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
+        open_kwargs.setdefault("mode", mode)
+        with zipfile.ZipFile(file=path) as zip_file:
+            with zip_file.open(inner_path) as file:
+                yield file
+
     @contextmanager
     def ensure_open_gz(
         self,
@@ -185,13 +347,15 @@ class Module:
         name: Optional[str] = None,
         force: bool = False,
         download_kwargs: Optional[Mapping[str, Any]] = None,
-        mode: str = 'rb',
+        mode: str = "rb",
         open_kwargs: Optional[Mapping[str, Any]] = None,
     ):
         """Ensure a gzipped file is downloaded then open it."""
-        path = self.ensure(*subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs)
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
         open_kwargs = {} if open_kwargs is None else dict(open_kwargs)
-        open_kwargs.setdefault('mode', mode)
+        open_kwargs.setdefault("mode", mode)
         with gzip.open(path, **open_kwargs) as file:
             yield file
 
@@ -203,13 +367,32 @@ class Module:
         force: bool = False,
         download_kwargs: Optional[Mapping[str, Any]] = None,
         read_csv_kwargs: Optional[Mapping[str, Any]] = None,
-    ) -> 'pd.DataFrame':
+    ) -> "pd.DataFrame":
         """Download a CSV and open as a dataframe with :mod:`pandas`."""
         import pandas as pd
 
-        path = self.ensure(*subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs)
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
         return pd.read_csv(path, **_clean_csv_kwargs(read_csv_kwargs))
 
+    def ensure_json(
+        self,
+        *subkeys: str,
+        url: str,
+        name: Optional[str] = None,
+        force: bool = False,
+        download_kwargs: Optional[Mapping[str, Any]] = None,
+        json_load_kwargs: Optional[Mapping[str, Any]] = None,
+    ):
+        """Download a JSON file and open it with :mod:`json`."""
+        with self.ensure_open(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        ) as file:
+            return json.load(file, **(json_load_kwargs or {}))
+
+    # TODO ensure_pickle
+
     def ensure_excel(
         self,
         *subkeys: str,
@@ -218,9 +401,11 @@ class Module:
         force: bool = False,
         download_kwargs: Optional[Mapping[str, Any]] = None,
         read_excel_kwargs: Optional[Mapping[str, Any]] = None,
-    ) -> 'pd.DataFrame':
+    ) -> "pd.DataFrame":
         """Download an excel file and open as a dataframe with :mod:`pandas`."""
-        path = self.ensure(*subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs)
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
         return pd.read_excel(path, **(read_excel_kwargs or {}))
 
     def ensure_tar_df(
@@ -230,13 +415,72 @@ class Module:
         inner_path: str,
         name: Optional[str] = None,
         force: bool = False,
-        sep: str = '\t',
         download_kwargs: Optional[Mapping[str, Any]] = None,
         read_csv_kwargs: Optional[Mapping[str, Any]] = None,
-    ) -> 'pd.DataFrame':
-        """Download a tar file and open an inner file as a dataframe with :mod:`pandas`."""
-        path = self.ensure(*subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs)
-        return read_tarfile_csv(path=path, inner_path=inner_path, sep=sep, **_clean_csv_kwargs(read_csv_kwargs))
+    ) -> "pd.DataFrame":
+        """Download a tar file and open an inner file as a dataframe with :mod:`pandas`.
+
+        :param subkeys:
+            A sequence of additional strings to join. If none are given,
+            returns the directory for this module.
+        :param url:
+            The URL to download.
+        :param inner_path:
+            The relative path to the file inside the archive
+        :param name:
+            Overrides the name of the file at the end of the URL, if given. Also
+            useful for URLs that don't have proper filenames with extensions.
+        :param force:
+            Should the download be done again, even if the path already exists?
+            Defaults to false.
+        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
+        :param read_csv_kwargs: Keyword arguments to pass through to :func:`pandas.read_csv`.
+        :returns: A dataframe
+
+        .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
+        """
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
+        return read_tarfile_csv(
+            path=path, inner_path=inner_path, **_clean_csv_kwargs(read_csv_kwargs)
+        )
+
+    def ensure_tar_xml(
+        self,
+        *subkeys: str,
+        url: str,
+        inner_path: str,
+        name: Optional[str] = None,
+        force: bool = False,
+        download_kwargs: Optional[Mapping[str, Any]] = None,
+        parse_kwargs: Optional[Mapping[str, Any]] = None,
+    ):
+        """Download a tar file and open an inner file as an XML with :mod:`lxml`.
+
+        :param subkeys:
+            A sequence of additional strings to join. If none are given,
+            returns the directory for this module.
+        :param url:
+            The URL to download.
+        :param inner_path:
+            The relative path to the file inside the archive
+        :param name:
+            Overrides the name of the file at the end of the URL, if given. Also
+            useful for URLs that don't have proper filenames with extensions.
+        :param force:
+            Should the download be done again, even if the path already exists?
+            Defaults to false.
+        :param download_kwargs: Keyword arguments to pass through to :func:`pystow.utils.download`.
+        :param parse_kwargs: Keyword arguments to pass through to :func:`lxml.etree.parse`.
+        :returns: An ElementTree object
+
+        .. warning:: If you have lots of files to read in the same archive, it's better just to unzip first.
+        """
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
+        return read_tarfile_xml(path=path, inner_path=inner_path, **(parse_kwargs or {}))
 
     def ensure_zip_df(
         self,
@@ -247,10 +491,14 @@ class Module:
         force: bool = False,
         download_kwargs: Optional[Mapping[str, Any]] = None,
         read_csv_kwargs: Optional[Mapping[str, Any]] = None,
-    ) -> 'pd.DataFrame':
+    ) -> "pd.DataFrame":
         """Download a zip file and open an inner file as a dataframe with :mod:`pandas`."""
-        path = self.ensure(*subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs)
-        return read_zipfile_csv(path=path, inner_path=inner_path, **_clean_csv_kwargs(read_csv_kwargs))
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
+        return read_zipfile_csv(
+            path=path, inner_path=inner_path, **_clean_csv_kwargs(read_csv_kwargs)
+        )
 
     def ensure_rdf(
         self,
@@ -261,19 +509,21 @@ class Module:
         download_kwargs: Optional[Mapping[str, Any]] = None,
         precache: bool = True,
         parse_kwargs: Optional[Mapping[str, Any]] = None,
-    ) -> 'rdflib.Graph':
+    ) -> "rdflib.Graph":
         """Download a RDF file and open with :mod:`rdflib`."""
-        path = self.ensure(*subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs)
+        path = self.ensure(
+            *subkeys, url=url, name=name, force=force, download_kwargs=download_kwargs
+        )
         if not precache:
             return read_rdf(path=path, **(parse_kwargs or {}))
 
-        cache_path = path.with_suffix(path.suffix + '.pickle.gz')
+        cache_path = path.with_suffix(path.suffix + ".pickle.gz")
         if cache_path.exists() and not force:
-            with gzip.open(cache_path, 'rb') as file:
+            with gzip.open(cache_path, "rb") as file:
                 return pickle.load(file)  # type: ignore
 
         rv = read_rdf(path=path, **(parse_kwargs or {}))
-        with gzip.open(cache_path, 'wb') as file:
+        with gzip.open(cache_path, "wb") as file:
             pickle.dump(rv, file, protocol=pickle.HIGHEST_PROTOCOL)  # type: ignore
         return rv
 
@@ -283,7 +533,7 @@ class Module:
         s3_bucket: str,
         s3_key: Union[str, Sequence[str]],
         name: Optional[str] = None,
-        client: Optional['botocore.client.BaseClient'] = None,
+        client: Optional["botocore.client.BaseClient"] = None,
         client_kwargs: Optional[Mapping[str, Any]] = None,
         download_file_kwargs: Optional[Mapping[str, Any]] = None,
         force: bool = False,
@@ -312,7 +562,7 @@ class Module:
             The path of the file that has been downloaded (or already exists)
         """
         if not isinstance(s3_key, str):
-            s3_key = '/'.join(s3_key)  # join sequence
+            s3_key = "/".join(s3_key)  # join sequence
         if name is None:
             name = name_from_s3_key(s3_key)
         path = self.join(*subkeys, name=name, ensure_exists=True)
@@ -359,5 +609,5 @@ class Module:
 def _clean_csv_kwargs(read_csv_kwargs):
     if read_csv_kwargs is None:
         read_csv_kwargs = {}
-    read_csv_kwargs.setdefault('sep', '\t')
+    read_csv_kwargs.setdefault("sep", "\t")
     return read_csv_kwargs
